NB. lex.fp - Why am I doing this?

NB.--token table------------------------------------------------------------

(ltable) = 
     @[unpack 1, 2]
       ~<<"...", DOTS>,
 	 <".", PERIOD>,
	 <"debug", DEBUG>,
	 <"where", WHERE>,
	 <"end", END>,
	 <"while", WHILE>,
	 <"bu", BU>,
	 <"catch", CATCH>,
	 <"->", ARROW>,
	 <";", SEMICOLON>,
	 <"<", OSEQUENCE>,
	 <">", CSEQUENCE>,
	 <"(", OPAREN>,
	 <")", CPAREN>,
	 <"[", OBRACKET>,
	 <"]", CBRACKET>,
	 <"{", OCURLY>,
	 <"}", CCURLY>,
	 <",", COMMA>,
	 <"=", EQUAL>,
	 <"#", SHARP>,
	 <"@", ALPHA>,
	 <"&", AND>,
	 <"|", OR>,
	 <":", APP>,
	 <"/", INSERT>,
	 <"\\", INSERTR>,
         <"`", EQO>,
	 <"*", LIFT>,
	 <"_", BOTTOM>,
	 <"verbatim", VERBATIM>,
 	 <"~", CONST> >.

NB.--Lexer entry point------------------------------------------------------

NB. lex:STRING -> <TOKEN, ...>
NB.   TOKEN = <TYPE, LINE, [VAL]>

lex = reverse lex'[id, ~1, 0] where

lex''{input=, line=, tokens=} =
  {} input -> tokens;
  {`10, more=...} input -> lex'[more input, succ line, tokens];
  {whitespace, more=...} input -> lex'[more input, line, tokens];
  {`$$, c=, more=...} input -> 
    lex'[more 1, line, apndl[[~CHAR, line, c 1], tokens]];
  {`$", more=...} input -> string apndl[more input, -1 state];
  {`$N | `$n, `$B | `$b, `$., more=...} input -> skip_line;
  lex_token.

lex_other{input=, line=, tokens=} =
  ({c=lowercase, more=...} input -> scan_id state;
   {c=uppercase, more=...} input -> scan_atom state;
   {c=digit, more=...} input -> scan_number state;
   {c=bu gt 32, more=...} input -> 
     lex'[more input, line, apndl[[~ID, line, pack[c input]], tokens]];
   lex_error[~"invalid character", line, 1 input]).

lex_token = find' apndl[ltable, id].

NB. lex' = lex'' debug "lex".
lex' = lex''.

NB.--Helper functions-------------------------------------------------------

find'{lt=, input=, line=, tokens=} =
  null lt  -> lex_other[input, line, tokens];
  prefix[1 1 lt, input] -> 
    lex'[drop[length 1 1 lt, input], line, apndl[[2 1 lt, line], tokens]];
  find'[-1 lt, input, line, tokens].
state = [1, 0, 2, 3].
whitespace = member[id, ~<32, 9, 10, 13>].
skip_line =
  {} 1 -> lex'[0, 2, 3];
  {`10, more=...} 1 -> lex'[more 1, succ 2, 3];
  skip_line[-1 1, 2, 3].
string{input=, str=, line=, tokens=} =
  {} input -> lex_error[~"illegal end of quoted atom literal", line];
  {`$", more=...} input -> 
    lex'[more input, line, apndl[[~ATOM, line, pack reverse str], tokens]];
  {`$\, c=, more=...} input ->
    string[more input, apndl[c input, str], line, tokens];
  string[-1 input, apndl[1 input, str], (`10) 1 input -> succ line; line, tokens].
scan_id{input=, str=, line=, tokens=} =
  (null | {not (alphanumeric | `$_ | `$'), ...}) input ->
    lex'[input, line, apndl[[~ID, line, pack reverse str], tokens]];
  scan_id[-1 input, apndl[1 input, str], line, tokens].
scan_atom{input=, str=, line=, tokens=} =
  (null | {not (alphanumeric | `$_), ...}) input ->
    lex'[input, line, apndl[[~ATOM, line, pack reverse str], tokens]];
  scan_atom[-1 input, apndl[1 input, str], line, tokens].
scan_number{input=, str=, line=, tokens=} =
  (null | {not digit, ...}) input ->
    lex'[input, line, apndl[[~NUM, line, number pack reverse str], tokens]];
  scan_number[-1 input, apndl[1 input, str], line, tokens].

NB.--Error reporting---------------------------------------------------------------

NB. lex_error:<msg, line, arg1, ...>

lex_error{msg=, line=, args=...} =
  throw[merge[~"lexer error in line ", line, ~": ", msg], args].

end.

NB.--Library stuff-----------------------------------------------------------------

NB. prefix:<charlist-prefix, charlist> -> bool

prefix = (le [1, length 3] -> eq [2, take[1, 3]]; ~F) [length 1, 1, 2].
lowercase = bu ge $a & bu le $z.
uppercase = bu ge $A & bu le $Z.
digit = bu ge $0 & bu le $9.
alphanumeric = digit | lowercase | uppercase.

NB.--------------------------------------------------------------------------------
NB.
NB. TODO:
NB.
NB. - some sort of progress output
NB.
